/*
    module/ni-E.c
    Hardware driver for NI E series cards

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
	This file needs to be included by another file, e.g.,
	atmio-E.c.

	Interrupt support added by Truxton Fulton <trux@truxton.com>

	References for specifications:
	
	   321747b.pdf  Register Level Programmer Manual (obsolete)
	   321747c.pdf  Register Level Programmer Manual (new)
	   DAQ-STC reference manual

	Other possibly relevant info:
	
	   320517c.pdf  User manual (obsolete)
	   320517f.pdf  User manual (new)
	   320889a.pdf  delete
	   320906c.pdf  maximum signal ratings
	   321066a.pdf  about 16x
	   321791a.pdf  discontinuation of at-mio-16e-10 rev. c
	   321808a.pdf  about at-mio-16e-10 rev P
	   321837a.pdf  discontinuation of at-mio-16de-10 rev d
	   321838a.pdf  about at-mio-16de-10 rev N
	
	ISSUES:

	need to deal with external reference for DAC, and other DAC
	properties in board properties
	
	deal with at-mio-16de-10 revision D to N changes, etc.
	
*/



static int ni_modebits1[4]={ 0x3000, 0x2000, 0x1000, 0 };
static int ni_modebits2[4]={ 0x3f, 0x3f, 0x37, 0x37 };

static int ni_gainlkup[][16]={
	{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 },
	{ 1, 2, 4, 7, 9, 10, 12, 15, 0,0,0,0,0,0,0,0 },
	{ 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 12, 13, 14, 15, 0,0 }
};

static int ni_n_ranges[]={ 16, 8, 14 };

static int ni_range_lkup[]={
	RANGE_ni_E_ai,
	RANGE_ni_E_ai_limited,
	RANGE_ni_E_ai_limited14
};



static int ni_dio(comedi_device *dev,comedi_trig *ad);
static int ni_ai(comedi_device *dev,comedi_trig *it);
static int ni_rem(comedi_device *dev);
static int ni_read_eeprom(comedi_device *dev,int addr);
static void ni_calibrate(comedi_device *dev);
static int ni_memcpy_from(comedi_device *dev,comedi_memcpy *it);


static void ni_E_interrupt(int irq,void *d,struct pt_regs * regs)
{
	comedi_sample s;
	comedi_device *dev=d;
	int wake=0 ;
	int status;

/*
    If you want to use windowed registers in an interrupt, it is
    important that you restore the window address register.  If
    you change certain modes, e.g., AI_Configuration_Start/End,
    you need to set up software flags for non-interrupt routines.
*/
	
	status=ni_readw(AI_Status_1+dev->iobase);
	if(status&(AI_Overrun_St|AI_Overflow_St))
		printk("ni_E: overrun/overflow status=0x%04x\n",status);
	
	if(status&AI_FIFO_Half_Full_St){
	/*
	   if we got a fifo_half_full interrupt, we can transfer fifo/2
	   samples witout checking the empty flag.  It doesn't matter if
	   we transfer the rest of the samples, the performance trade-off
	   is minimal (checking empty flag for a few samples vs. having
	   1% more interrupts.)  At really high speeds, it's better to
	   ignore them.
	*/
		int i;
	
#if 0
printk("fifo half full\n");
#endif
		for(i=0;i<boardtype.ai_fifo_depth/2;i++){
			s.data=ni_readw(dev->iobase+ADC_FIFO_Data_Register);
			if(!(devpriv->aip[devpriv->lastchan]&0x100))
				s.data^=0x800;
			s.data&=0xfff; /* incompatible with 16 bit cards */
			s.chan=devpriv->lastchan;
			report_sample(dev,&s);
		}
		return;
	}
	if(status&AI_SC_TC_St){
	/* 
	    end of sampling.  Read the rest of the FIFO.
	*/
#if 0
printk("terminal count\n");
#endif
		while(!(ni_readw(AI_Status_1+dev->iobase)&AI_FIFO_Empty_St)){
			s.data=ni_readw(dev->iobase+ADC_FIFO_Data_Register);
			if(!(devpriv->aip[devpriv->lastchan]&0x100))
				s.data^=0x800;
			s.data&=0xfff; /* incompatible with 16 bit cards */
			s.chan=devpriv->lastchan;
			report_sample(dev,&s);
			wake=1;
		}
		dev->busy[0]=0;
	}
	if(wake){
		wake_up_interruptible(&dev->rdwait);
	}
}

#if 0
static int ni_stop_job(comedi_device *dev,int job)
{
/* assume it means ai job*/
	win_out(AI_Configuration_Start,Joint_Reset_Register);
	win_out(AI_Configuration_End,Joint_Reset_Register);

	return 0;
}
#endif

static int ni_iai(comedi_device *dev,comedi_trig *it)
{
	int mode,chan;
	int i;
	
#ifdef TRUX
	/* this is a hack, but then, so is busy[], so I'll allow it for a while   -ds */
	/* special case : kill job */
	if(it->trigsrc==COMEDI_TRIGRESET)
	  {
	    ni_stop_job(dev,dev->busy[0]) ;
	    dev->busy[0]=0 ;
	    return(0) ;
	  }

#endif /* TRUX */

	if(dev->busy[0])return -EBUSY;
	
	win_out(1,ADC_FIFO_Clear);
	win_out(1,Configuration_Memory_Clear);

	/* interrupt on errors */
	win_out(0x0020,Interrupt_A_Enable_Register) ;
	
	mode=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_AREF);
	chan=ni_modebits1[mode]|(it->chan&ni_modebits2[mode]);

	ni_writew(chan,dev->iobase+Configuration_Memory_High);
	ni_writew(0x8000|devpriv->aip[it->chan],
		dev->iobase+Configuration_Memory_Low);
	win_out(1,AI_Command_1_Register);
	for(i=0;i<40;i++){
		if(!(ni_readw(AI_Status_1+dev->iobase)&
		  AI_FIFO_Empty_St)){
			win_out(1,ADC_FIFO_Clear);
			break;
		}
		udelay(25);
	}
	if(i==40){
		printk("ni_E: timeout 1\n");
		return -ETIME;
	}

	win_out(1,AI_Command_1_Register);
	
	for(i=0;i<40;i++){
		if(!(ni_readw(AI_Status_1+dev->iobase)&
		  AI_FIFO_Empty_St)){
			it->data=ni_readw(dev->iobase+ADC_FIFO_Data_Register);
			if(boardtype.adbits==12){
				if(!(devpriv->aip[it->chan]&0x100))
					it->data^=0x800;
				it->data&=0xfff;
			}else{
				if(!(devpriv->aip[it->chan]&0x100))
					it->data^=0x8000;
				it->data&=0xffff;
			}
			return 0;
		}
		udelay(25);
	}
	printk("ni_E: timeout 2\n");
	return -ETIME;
}


static int ni_mtrig(comedi_device *dev,comedi_mtrig *it,unsigned int *chanlist)
{
	int mode,chan;
	int i;
	
	switch(it->io){
	case COMEDI_IO_INPUT:
	{
		/* use FIFO interrupts to clean FIFO */
		devpriv->lastchan=chanlist[0];
		
		win_out(1,ADC_FIFO_Clear);
		win_out(1,Configuration_Memory_Clear);
	
		for(i=0;i<it->n1;i++){
			mode=getparamval(dev->chinfo[chanlist[i]].paramlist,COMEDI_AREF);

			chan=ni_modebits1[mode]|(chanlist[i]&ni_modebits2[mode]);
			ni_writew(chan,dev->iobase+Configuration_Memory_High);

			ni_writew(((i==it->n1-1)?0x8000:0)|devpriv->aip[chanlist[i]],
				dev->iobase+Configuration_Memory_Low);
		}
		
		/* start configuration */
		win_out(AI_Configuration_Start,Joint_Reset_Register);
	
		/* stage number of samples */
		win_out((it->n2-1)>>16,AI_SC_Load_A_Registers);
		win_out((it->n2-1)&0xffff,AI_SC_Load_A_Registers+1);
		
		/* load SC */
		win_out(0x20,AI_Command_1_Register);
		
		/*
		    AI_SI_Special_Trigger_Delay=0
		    AI_Pre_Trigger=0
		  AI_START_STOP_Select_Register:
		    AI_START_Polarity=0 (?)	rising edge
		    AI_START_Edge=1		edge triggered
		    AI_START_Sync=1 (?)		
		    AI_START_Select=0		SI_TC
		    AI_STOP_Polarity=0		rising edge
		    AI_STOP_Edge=0		level
		    AI_STOP_Sync=1		
		    AI_STOP_Select=19		external pin (configuration mem)
		win_out(AI_START_Edge|AI_START_Sync|AI_STOP_Select(19)|
			AI_STOP_Sync,AI_START_STOP_Register);
		*/
		win_out((it->trigvar2>>16),AI_SI_Load_A_Registers);
		win_out((it->trigvar2&0xffff),AI_SI_Load_A_Registers+1);
		/* AI_SI_Initial_Load_Source=A */
		win_out(0,AI_Mode_2_Register);
		/* load SI */
		win_out(0x200,AI_Command_1_Register);

		/* stage freq. counter into SI B */
		win_out((it->trigvar2>>16),AI_SI_Load_B_Registers);
		win_out((it->trigvar2&0xffff),AI_SI_Load_B_Registers+1);
	
		win_out(it->trigvar1,AI_SI2_Load_A_Register); /* 0,0 does not work. */
		win_out(it->trigvar1,AI_SI2_Load_B_Register);

		/* AI_SI2_Reload_Mode = alternate */
		/* AI_SI2_Initial_Load_Source = A */
		win_out(0x0100,AI_Mode_2_Register);
	
		/* AI_SI2_Load */
		win_out(0x0800,AI_Command_1_Register);
	
		/* AI_SI_Initial_Load_Source=0
		   AI_SI_Reload_Mode(0)
		   AI_SI2_Reload_Mode = alternate, AI_SI2_Initial_Load_Source = B */
		win_out(0x0300,AI_Mode_2_Register);

		/* end configuration */
		win_out(AI_Configuration_End,Joint_Reset_Register);

		if(dev->irq){
			/* interrupt on FIFO, errors, SC_TC */
			win_out(0x00a1,Interrupt_A_Enable_Register) ;
		}else{
			/* interrupt on nothing */
			win_out(0x0000,Interrupt_A_Enable_Register) ;
		}
		
		/* AI_SI2_Arm, AI_SI_Arm, AI_DIV_Arm, AI_SC_Arm */
		win_out(0x1540,AI_Command_1_Register);

		/* AI_START1_Pulse */
		win_out(1,AI_Command_2_Register);

		dev->busy[0]=it->job;
		if(!dev->irq){
			start_polling(dev);
		}
		return 0;
	}
	default:
		return -EINVAL;
	}
}

static int ni_ai(comedi_device *dev,comedi_trig *it)
{
	int mode,chan;
	
	if(dev->busy[0])return -EBUSY;
	switch(it->trigsrc){
	case COMEDI_TRIGCLK:
		/* use FIFO interrupts to clean FIFO */
		devpriv->lastchan=it->chan;
		
		win_out(1,ADC_FIFO_Clear);
		win_out(1,Configuration_Memory_Clear);
	
		mode=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_AREF);
		chan=ni_modebits1[mode]|(it->chan&ni_modebits2[mode]);
	
		ni_writew(chan,dev->iobase+Configuration_Memory_High);
		ni_writew(0x8000|devpriv->aip[it->chan],
			dev->iobase+Configuration_Memory_Low);
		
		/* start configuration */
		win_out(AI_Configuration_Start,Joint_Reset_Register);
	
		/* stage number of samples */
		win_out((it->n-1)>>16,AI_SC_Load_A_Registers);
		win_out((it->n-1)&0xffff,AI_SC_Load_A_Registers+1);
		
		/* load SC */
		win_out(0x20,AI_Command_1_Register);
		
		/*
		    AI_SI_Special_Trigger_Delay=0
		    AI_Pre_Trigger=0
		  AI_START_STOP_Select_Register:
		    AI_START_Polarity=0 (?)	rising edge
		    AI_START_Edge=1		edge triggered
		    AI_START_Sync=1 (?)		
		    AI_START_Select=0		SI_TC
		    AI_STOP_Polarity=0		rising edge
		    AI_STOP_Edge=0		level
		    AI_STOP_Sync=1		
		    AI_STOP_Select=19		external pin (configuration mem)
		win_out(AI_START_Edge|AI_START_Sync|AI_STOP_Select(19)|
			AI_STOP_Sync,AI_START_STOP_Register);
		*/

		win_out((it->trigvar>>16),AI_SI_Load_A_Registers);
		win_out((it->trigvar&0xffff),AI_SI_Load_A_Registers+1);
		/* AI_SI_Initial_Load_Source=A */
		win_out(0,AI_Mode_2_Register);
		/* load SI */
		win_out(0x200,AI_Command_1_Register);

		/* stage freq. counter into SI B */
		win_out((it->trigvar>>16),AI_SI_Load_B_Registers);
		win_out((it->trigvar&0xffff),AI_SI_Load_B_Registers+1);
	
		win_out(1,AI_SI2_Load_A_Register); /* 0,0 does not work. */
		win_out(1,AI_SI2_Load_B_Register);

		/* AI_SI2_Reload_Mode = alternate */
		/* AI_SI2_Initial_Load_Source = A */
		win_out(0x0100,AI_Mode_2_Register);
	
		/* AI_SI2_Load */
		win_out(0x0800,AI_Command_1_Register);
	
		/* AI_SI_Initial_Load_Source=0
		   AI_SI_Reload_Mode(0)
		   AI_SI2_Reload_Mode = alternate, AI_SI2_Initial_Load_Source = B */
		win_out(0x0300,AI_Mode_2_Register);

		/* end configuration */
		win_out(AI_Configuration_End,Joint_Reset_Register);

		if(dev->irq){
			/* interrupt on FIFO, errors, SC_TC */
			win_out(0x00a1,Interrupt_A_Enable_Register) ;
		}else{
			/* interrupt on nothing */
			win_out(0x0000,Interrupt_A_Enable_Register) ;
		}
		
		/* AI_SI2_Arm, AI_SI_Arm, AI_DIV_Arm, AI_SC_Arm */
		win_out(0x1540,AI_Command_1_Register);

		/* AI_START1_Pulse */
		win_out(1,AI_Command_2_Register);

		dev->busy[0]=it->job;
		if(!dev->irq){
			start_polling(dev);
		}
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
	{
		int pval;
		
		if(it->pval<0 || it->pval>=ni_n_ranges[boardtype.gainlkup])return -EINVAL;

		pval=ni_gainlkup[boardtype.gainlkup][it->pval];
		
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		devpriv->aip[it->chan]&=~0x107;
		devpriv->aip[it->chan]|=((pval&0x8)<<5);
		devpriv->aip[it->chan]|=(pval&0x7);
		return 0;
	}
	case COMEDI_AREF:
		if(it->pval<0 || it->pval>=4)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_AREF,
			0,it->pval);
		return 0;
	case COMEDI_DITHER:
		if(boardtype.alwaysdither)return -EINVAL;
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_DITHER,
			0,it->pval);
		devpriv->aip[it->chan]&=~0x200;
		devpriv->aip[it->chan]|=(it->pval<<9);
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_ao(comedi_device *dev,comedi_trig *it)
{
	unsigned int i=it->data;
	
	if(i>=0x1000)return -EINVAL;
	i^=0x800;
	if(it->chan&1){
		if(!(devpriv->ao0p&1))i^=0x800;
		ni_writew(i,dev->iobase+DAC0_Direct_Data);
	}else{
		if(!(devpriv->ao1p&1))i^=0x800;
		ni_writew(i,dev->iobase+DAC1_Direct_Data);
	}
	return 0;
}

static int ni_aosp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		if(it->chan==16){
			devpriv->ao0p&=~1;
			devpriv->ao0p|=(it->pval?1:0);
			ni_writew(devpriv->ao0p,dev->iobase+AO_Configuration);
		}else{
			devpriv->ao1p&=~1;
			devpriv->ao1p|=(it->pval?1:0);
			ni_writew(devpriv->ao1p,dev->iobase+AO_Configuration);
		}
		return 0;
	case COMEDI_DEGLITCH:
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_DEGLITCH,
			0,it->pval);
		if(it->chan==16){
			devpriv->ao0p&=~2;
			devpriv->ao0p|=(it->pval?2:0);
			ni_writew(devpriv->ao0p,dev->iobase+AO_Configuration);
		}else{
			devpriv->ao1p&=~2;
			devpriv->ao1p|=(it->pval?2:0);
			ni_writew(devpriv->ao1p,dev->iobase+AO_Configuration);
		}
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_dio(comedi_device *dev,comedi_trig *it)
{
	int data,mask;

	switch(it->io){
	case COMEDI_IO_INPUT:
		it->data= win_in(DIO_Input_Register)&0xff;
	
		return 0;
	case COMEDI_IO_OUTPUT:
		if(it->n)mask=it->n; else mask=0xff;
		data=(devpriv->last_do&~mask)|(it->data&mask);
		win_out(data,DIO_Output_Register);
		devpriv->last_do=data;
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_dsp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_IOBITS:
		if(it->pval&~0xff)return -EINVAL;
		devpriv->dio=it->pval;
		win_out(devpriv->dio,DIO_Control_Register);
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_IOBITS,0,
			it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

/*
   XXX this assumes that if there isn't anything in the fifo,
   then we can turn off polling.  this is *not* a good idea,
   but it will do for now.
   Note: there is always the possibility that the poll
   routine will steal a sample from a soft triggered AI.
*/

static void ni_E_poll(comedi_device *dev)
{
	comedi_sample s;
	int wake=0;
	int i;
	int status;
#ifdef DEBUG
	int n=0;
#endif
	
	status=ni_readw(AI_Status_1+dev->iobase);
#ifdef DEBUG
	printk("ni_E: polling... jiffies=%d status=0x%04x ",jiffies,status);
#endif

	if(status&AI_FIFO_Half_Full_St){
#ifdef DEBUG
		printk("fifo half full ");
		n+=boardtype.ai_fifo_depth/2;
#endif
		for(i=0;i<boardtype.ai_fifo_depth/2;i++){
			s.data=ni_readw(dev->iobase+ADC_FIFO_Data_Register);
			if(!(devpriv->aip[devpriv->lastchan]&0x100))
				s.data^=0x800;
			s.data&=0xfff; /* incompatible with 16 bit cards */
			s.chan=devpriv->lastchan;
			report_sample(dev,&s);
		}
		wake=1;
	}
	while(!(ni_readw(AI_Status_1+dev->iobase)&AI_FIFO_Empty_St)){
		s.data=ni_readw(dev->iobase+ADC_FIFO_Data_Register);
		if(!(devpriv->aip[devpriv->lastchan]&0x100))
			s.data^=0x800;
		s.data&=0xfff; /* incompatible with 16 bit cards */
		s.chan=devpriv->lastchan;
		report_sample(dev,&s);
		wake=1;
#ifdef DEBUG
		n++;
#endif
	}
#ifdef DEBUG
	printk("samples=%d\n",n);
#endif
	if(wake)wake_up_interruptible(&dev->rdwait);
	else{
		stop_polling(dev);
		dev->busy[0]=0;
	}
}



static int ni_E_init(comedi_device *dev,comedi_devconfig *it,int board)
{
	int i;

	devpriv=kmalloc(sizeof(ni_private)+dev->chan*sizeof(comedi_chinfo),
		GFP_KERNEL);
	if(!devpriv){
		return -ENOMEM;
	}
	memset(devpriv,0,sizeof(ni_private)+dev->chan*sizeof(comedi_chinfo));
	devpriv->board=board;
	
	dev->chinfo=(void *)(devpriv+1);
	for(i=0;i<boardtype.n_adchan;i++){
		dev->chinfo[i].itrig=ni_iai;
		dev->chinfo[i].trig=ni_ai;
		dev->chinfo[i].sp=ni_aisp;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,boardtype.adbits);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,(1<<boardtype.adbits)-1);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
			COMEDI_RDONLY,ni_range_lkup[boardtype.gainlkup]);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AREF,COMEDI_RDWR,
			COMEDI_AREF_DIFF);
		if(boardtype.alwaysdither){
			addparam(&dev->chinfo[i].paramlist,COMEDI_DITHER,
				COMEDI_RDONLY,1);
			devpriv->aip[i]=0x0100;
		}else{
			addparam(&dev->chinfo[i].paramlist,COMEDI_DITHER,
				COMEDI_RDWR,0);
			devpriv->aip[i]=0x0000;
		}
	}
	dev->chinfo[i].itrig=ni_ao;
	dev->chinfo[i].sp=ni_aosp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,12);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_MAXDATA,COMEDI_RDONLY,(1<<12)-1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
		COMEDI_RDONLY,RANGE_bipolar10);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	addparam(&dev->chinfo[i].paramlist,COMEDI_DEGLITCH,0,0);
	i++;
	
	dev->chinfo[i].itrig=ni_ao;
	dev->chinfo[i].sp=ni_aosp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,12);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
		COMEDI_RDONLY,RANGE_bipolar10);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	addparam(&dev->chinfo[i].paramlist,COMEDI_DEGLITCH,0,0);
	i++;

	dev->chinfo[i].itrig=ni_dio;
	dev->chinfo[i].sp=ni_dsp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_BIDIR);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITGROUP,COMEDI_RDONLY,1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITS,COMEDI_RDWR,0xff&COMEDI_IOBITS_INPUT);
	devpriv->dio=0;
	win_out(devpriv->dio,DIO_Control_Register);

	win_out(AI_Configuration_Start,Joint_Reset_Register);

	win_out(0x3f80,Interrupt_A_Ack_Register); /* clear interrupts */
	win_out(0x0000,AI_Command_1_Register); /* reset pulses */
	win_out(0x1ba0,Clock_and_FOUT_Register);
	win_out(0x000d,AI_Mode_1_Register);
	win_out(0x0000,AI_Mode_2_Register);
	win_out((1<<6)|0x0000,AI_Mode_3_Register); /* generate FIFO interrupts on half full */
	win_out(0xa4a0,AI_Personal_Register); /* ? */
	win_out(0x032e,AI_Output_Control_Register);
	win_out(0x0060,AI_Trigger_Select_Register); /* trigger source */
	win_out(0x29e0,AI_START_STOP_Select_Register);

	win_out(AI_Configuration_End,Joint_Reset_Register);
	
	/* the following registers should not be changed:
		Clock_and_FOUT_Register
		AI_Mode_1_Register
		AI_Mode_3_Register
		AI_Personal_Register
		AI_Output_Control_Register
		AI_Trigger_Select_Register
		AI_START_STOP_Select_Register
	*/

	
	devpriv->ao0p=0x0000;
	ni_writew(devpriv->ao0p,dev->iobase+AO_Configuration);
	devpriv->ao1p=0x0100;
	ni_writew(devpriv->ao1p,dev->iobase+AO_Configuration);
	win_out(AO_Configuration_Start,Joint_Reset_Register);
	win_out(0x2000,AO_Command_1_Register);
	win_out(0,Interrupt_B_Enable_Register);
	win_out(0x0010,AO_Personal_Register);
	ni_writew(0x3f98,dev->iobase+Interrupt_B_Ack);
	win_out(0x1430,AO_Personal_Register);
	win_out(0,AO_Output_Control_Register);
	win_out(0,AO_Start_Select_Register);
	win_out(0,AO_Command_1_Register);

#if ATMIO
	ni_calibrate(dev);
#endif
	
	comedi_initbuf(&dev->rdbuf);

	dev->poll=ni_E_poll;
	dev->memcpy_from=ni_memcpy_from;
	dev->rem=ni_rem;
	dev->mtrig=ni_mtrig;

	printk("\n");

	return 1;
}


static int ni_rem(comedi_device *dev)
{
	int i;

	printk("comedi%d: ni_E: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(devpriv);
	release_region(dev->iobase,dev->iosize);
	if(dev->irq)free_irq(dev->irq,dev);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}


static int ni_read_eeprom(comedi_device *dev,int addr)
{
	int bit;
	int bitstring;

	bitstring=0x0300|((addr&0x100)<<3)|(addr&0xff);
	ni_writeb_p(0x04,dev->iobase+Serial_Command);
	for(bit=0x8000;bit;bit>>=1){
		ni_writeb_p(0x04|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
		ni_writeb_p(0x05|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
	}
	bitstring=0;
	for(bit=0x80;bit;bit>>=1){
		ni_writeb_p(0x04,dev->iobase+Serial_Command);
		ni_writeb_p(0x05,dev->iobase+Serial_Command);
		bitstring|=((ni_readb_p(dev->iobase+Status)&0x01)?bit:0);
	}
	ni_writeb_p(0x00,dev->iobase+Serial_Command);
	
	return bitstring;
}

static void ni_write_caldac(comedi_device *dev,int addr,int val)
{
	int loadbit,bits,bit,bitstring;
	
	switch(boardtype.caldactype){
	case 1:
		/*
		   MB 88341
		   Note that address bits are reversed.  Thanks to
		   Ingo Keen for noticing this.
		*/
		bits=12;
		bitstring=((addr&0x1)<<11) |
			  ((addr&0x2)<<9)  |
			  ((addr&0x4)<<7)  |
			  ((addr&0x8)<<5)  |
			  (val&0xff);
		loadbit=0x08;
		break;
	case 2:
		if(addr==8){
			/*
			   DAC 8043
			*/
			bits=12;
			bitstring=val&0xfff;
			loadbit=0x10;
		}else{
			/*
			   DAC 8800
			*/
			bits=11;
			bitstring=((addr&0x7)<<8)|(val&0xff);
			loadbit=0x08;
		}
		break;
	default:
		return;
	}

	for(bit=1<<(bits-1);bit;bit>>=1){
		ni_writeb(((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
		ni_writeb(1|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
	}
	ni_writeb(loadbit,dev->iobase+Serial_Command);
	ni_writeb(0,dev->iobase+Serial_Command);
}


static void ni_calibrate(comedi_device *dev)
{
	switch(boardtype.eepromlayout){
	case 1:
		ni_write_caldac(dev,4,ni_read_eeprom(dev,424));
		ni_write_caldac(dev,1,ni_read_eeprom(dev,423));
		ni_write_caldac(dev,3,ni_read_eeprom(dev,422));
		ni_write_caldac(dev,2,ni_read_eeprom(dev,421));
		break;
	case 2:
		ni_write_caldac(dev,4,ni_read_eeprom(dev,422));
		ni_write_caldac(dev,11,ni_read_eeprom(dev,421));
		ni_write_caldac(dev,1,ni_read_eeprom(dev,420));
		ni_write_caldac(dev,3,ni_read_eeprom(dev,419));
		ni_write_caldac(dev,2,ni_read_eeprom(dev,418));
		break;
	default:
	}
}

static int ni_memcpy_from(comedi_device *dev,comedi_memcpy *it)
{
	char *l=it->local;
	int i;
	
	if(it->remote<0 || it->nbytes<0 ||
	  it->remote+it->nbytes>512)return -EINVAL;
	for(i=0;i<it->nbytes;i++){
		l[i]=ni_read_eeprom(dev,it->remote+i);
	}
	return 0;
}

/*
 *
 *  Programmable Function Inputs
 *
 */

#if 0

static void pfi_setup(comedi_device *dev)
{
	/* currently, we don't output any signals, thus, all
	   the PFI's are input */
	
	win_out(IO_Bidirection_Pin_Register,0);
}



/*
 *
 *  General Purpose Counter/Timer section
 *
 */


/* event counting */

int gpct_start(comedi_device *dev,int chan)
{
	/* select load source */
	Gi_Load_Source_Select = 0;
	
	/* load initial counter value */
	Gi_Load_A = 0;
	
	/* strobe */
	Gi_Load=1;

/* command reg */
	Gi_Up_Down = 1; /* up counting */
	Gi_Bank_Switch_Enable = 0;
	Gi_Bank_Switch_Mode = 0;

/* input select reg */
	Gi_Source_Select = PFIn;
	Gi_Source_Polarity = 0; /* rising edges */
	Gi_Gate_Select = 31; /* logic low */
	Gi_OR_Gate = 0;
	Gi_Output_Polarity = 1; /* active low */
	Gi_Gate_Select_Load_Source = 0;

/* mode reg */
	Gi_Gate_Polarity = 0; /* disable inversion */
	Gi_Loading_On_Gate = 0;
	Gi_Output_Mode = 1; /* one clock cycle output */
	Gi_Loading_On_TC = 0;
	Gi_Gating_Mode = 1;
	Gi_Gate_On_Both_Edges = 0;
	Gi_Trigger_Mode_For_Edge_Gate = 2;
	Gi_Stop_Mode = 0;
	Gi_Counting_Once = 0;
	
	
/* int enable reg -- ignore */
	
/*
	Gi_TC_Interrupt_Enable = 0;
	Gi_Gate_Interrupt_Enable = 0;
*/
}

static int gpct_sp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_CTR_INPUT:
	{
		/* which input source? */
		int pval=it->pval;
		
		if(pval<0 || pval>=10)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_CTR_INPUT,
			0,pval);
		/* XXX set it */
		return 0;
	}
	case COMEDI_CTR_POLARITY:
		/* count rising or falling edges? */
		if(it->pval<0 || it->pval>=4)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_AREF,
			0,it->pval);
		/* XXX set it */
		return 0;
	case COMEDI_COUNT:
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_COUNT,
			0,it->pval);  /* irrelevant... */
		devpriv->gpct[it->chan-boardtype->counteroffset]=it->pval;
		return 0;
	}
	return -EINVAL;
}

#endif

